/*============================================================================
  File:     MultipurposeProcedures_Solution3.sql

  Summary:  Because OPTION(RECOMPILE) has such a HORRIBLE history
            this solution is really the best. However, it requires
            dynamically constructed strings so you'll want to make
            sure that you add a level of security / prevention for
            SQL injection.
  
  SQL Server Version: 2008+
------------------------------------------------------------------------------
  Written by Kimberly L. Tripp, SYSolutions, Inc.

  For more scripts and sample code, check out http://www.SQLskills.com

  This script is intended only as a supplement to demos and lectures
  given by Kimberly L. Tripp.  
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

USE [Credit];
GO

-----------------------------------------------------
-- Building the ultimate multipurpose procedure
-----------------------------------------------------

-- Best Option - is a hybrid combination of sp_executesql
--               and EXEC (@spexecutesqlStr)

-- Also, to reduce SQLinjection, check out this blog post
-- http://www.sqlskills.com/blogs/kimberly/little-bobby-tables-sql-injection-and-execute-as/

IF OBJECTPROPERTY (OBJECT_ID (N'dbo.GetMemberInformation')
    , N'IsProcedure') = 1
	DROP PROCEDURE [dbo].[GetMemberInformation];
GO

IF USER_ID('User_GetMemberInformation') IS NOT NULL
    DROP USER [User_GetMemberInformation];
GO

CREATE USER [User_GetMemberInformation]
WITHOUT LOGIN;
GO
 
GRANT SELECT ON [dbo].[Member] 
TO [User_GetMemberInformation];
GO

-- You'll need this if you want to review the showplan
-- for these executions.
GRANT SHOWPLAN TO [User_GetMemberInformation];
GO

CREATE PROC [dbo].[GetMemberInformation]
(
    @member_no	INT = NULL
    , @Lastname	VARCHAR (15) = NULL
	, @Firstname	VARCHAR (15) = NULL
    , @MiddleInitial  letter = NULL
    , @EmailAddress VARCHAR(128) = NULL
    , @Region_no  numeric_id = NULL
    , @Member_code status_code = NULL
)
WITH EXECUTE AS N'User_GetMemberInformation'
AS
IF (@member_no IS NULL
    AND @Lastname IS NULL
	AND @Firstname IS NULL
    AND @MiddleInitial IS NULL
    AND @EmailAddress IS NULL
    AND @Region_no IS NULL
    AND @Member_code IS NULL)
BEGIN
	RAISERROR ('You must supply at least one parameter.', 16, -1);
    RETURN;
END;

DECLARE @spexecutesqlStr	NVARCHAR (4000),
        @ExecStr            NVARCHAR (4000),
        @Recompile          BIT = 1;

SELECT @spexecutesqlStr =
	N'SELECT [m].* FROM [dbo].[member] AS [m] WHERE 1=1';

SELECT @ExecStr =
	N'SELECT [m].* FROM [dbo].[member] AS [m] WHERE 1=1';

IF @member_no IS NOT NULL
BEGIN
	SELECT @spexecutesqlStr = @spexecutesqlStr 
		+ N' AND [m].[member_no] = @MemberNo';
	SELECT @ExecStr = @ExecStr 
		+ N' AND [m].[member_no] = CONVERT(INT, ' + CONVERT(NVARCHAR(30),@member_no) + N')';
END

IF @LastName IS NOT NULL
BEGIN
	SELECT @spexecutesqlStr = @spexecutesqlStr 
		+ N' AND [m].[lastname] LIKE @LName'; 
    SELECT @ExecStr = @ExecStr 
		+ N' AND [m].[lastname] LIKE CONVERT(VARCHAR(15),' + QUOTENAME(@LastName, '''''') + N')';
END

IF @FirstName IS NOT NULL
BEGIN
	SELECT @spexecutesqlStr = @spexecutesqlStr 
		+ N' AND [m].[firstname] LIKE @FName';
    SELECT @ExecStr = @ExecStr 
		+ N' AND [m].[firstname] LIKE CONVERT(VARCHAR(15),' + QUOTENAME(@FirstName, '''''') + N')';
END

IF @MiddleInitial IS NOT NULL
BEGIN
	SELECT @spexecutesqlStr = @spexecutesqlStr 
		+ N' AND [m].[middleinitial] = @MI';
    SELECT @ExecStr = @ExecStr 
		+ N' AND [m].[middleinitial] = CONVERT(CHAR(1), ' + QUOTENAME(@MiddleInitial, '''''') + N')';
END

IF @EmailAddress IS NOT NULL
BEGIN
	SELECT @spexecutesqlStr = @spexecutesqlStr 
		+ N' AND [m].[email] LIKE @Email';
    SELECT @ExecStr = @ExecStr 
		+ N' AND [m].[email] LIKE CONVERT(VARCHAR(128),' + QUOTENAME(@EmailAddress, '''''') + N')';
END

IF @Region_no IS NOT NULL
BEGIN
	SELECT @spexecutesqlStr = @spexecutesqlStr 
		+ N' AND [m].[region_no] = @RegionNo';
    SELECT @ExecStr = @ExecStr 
		+ N' AND [m].[region_no] = CONVERT(TINYINT, ' + CONVERT(NVARCHAR(5),@Region_no) + N')';
END

IF @Member_code IS NOT NULL
BEGIN
	SELECT @spexecutesqlStr = @spexecutesqlStr 
		+ N' AND [m].[member_code] = @MemberCode';
    SELECT @ExecStr = @ExecStr 
		+ N' AND [m].[member_code] = CONVERT(CHAR(2), ' + QUOTENAME(@Member_code, '''''') + N')';
END

-- Conditions that you know will be selective and therefore
-- don't need to be recompiled.        
IF (@member_no IS NOT NULL)
    SET @Recompile = 0

IF (PATINDEX('%[%_?]%', @LastName) >= 4 
        OR PATINDEX('%[%_?]%', @LastName) = 0)
    AND (PATINDEX('%[%_?]%', @FirstName) >= 4
        OR PATINDEX('%[%_?]%', @FirstName) = 0)
    SET @Recompile = 0

IF (PATINDEX('%[%_?]%', @EmailAddress) >= 4 
        OR PATINDEX('%[%_?]%', @EmailAddress) = 0)
    SET @Recompile = 0

IF @Recompile = 1
BEGIN
    -- Use this next line for testing
    -- SELECT @ExecStr -- For testing
    EXEC (@ExecStr);
END
ELSE    
BEGIN
    -- Use this next line for testing
    -- SELECT @spexecutesqlStr, @Lastname, @Firstname, @CustomerID;
    EXEC [sp_executesql] @spexecutesqlStr
    , N'@MemberNo int, @LName varchar(15), @FName varchar(15)
    , @MI letter, @Email nvarchar(128), @RegionNo numeric_id
    , @MemberCode status_code'
	, @MemberNo = @Member_no
    , @LName = @Lastname
	, @FName = @Firstname
    , @MI = @MiddleInitial
    , @Email = @EmailAddress
    , @RegionNo = @Region_no
    , @MemberCode = @Member_code;
END
GO

SET STATISTICS IO ON;
GO

-- Let's try a couple of executions
EXEC [dbo].[GetMemberInformation] 
    @Member_no = 9912;
GO

EXEC [dbo].[GetMemberInformation] 
    @Lastname = 'Tripp'
    , @Firstname = 'Kimberly';
GO

EXEC [dbo].[GetMemberInformation] 
    @EmailAddress = N'BFXETPSUKOP.X.FLORINI@company02.com';
GO

EXEC [dbo].[GetMemberInformation] 
    @EmailAddress = '%2%.com'
    , @Region_No = 6
    , @Member_Code = 2;
GO  

EXEC [dbo].[GetMemberInformation] 
    @Member_no = 1234;
GO

EXEC [dbo].[GetMemberInformation] 
    @EmailAddress = N'GXRERSSCFTMISY.R.VANN@company16.com';
GO

EXEC [dbo].[GetMemberInformation] 
    @EmailAddress = '%27.com';
GO

EXEC [dbo].[GetMemberInformation] 
    @Lastname = 'Ran%'
    , @Firstname = 'P%';
GO  

EXEC [dbo].[GetMemberInformation] 
    @EmailAddress = N'NWGUMXWETJPPSN.G.JONES@company01.com';
GO

EXEC [dbo].[GetMemberInformation] 
    @Member_no = 2479;
GO

EXEC [dbo].[GetMemberInformation] 
    @EmailAddress = '%2%.com'
    , @Region_No = 6
    , @Lastname = 'L%'
    , @Member_Code = 2;
GO  

EXEC [dbo].[GetMemberInformation] 
    @EmailAddress = N'PZJWXGMEXPTSQJ.J.EFLIN@company16.com';
GO

sp_help status_code